/* Copyright (c) 1997-2012
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Darmstadt, Germany)
   http://www.polymake.org

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
--------------------------------------------------------------------------------
   $Project: polymake $$Id: minkowski_sum_fukuda.cc 10638 2012-03-14 13:15:29Z schroeter $*/

// This code was originally written by Silke Moeser in 2011.
// It was later modified by Benjamin Schroeter and others.

//TODO benutze datentyp ListMatrix

#include "polymake/polytope/minkowski_sum_fukuda.h"

namespace polymake { namespace polytope {

    namespace {
      typedef hash_set<Vector<Rational> > vertex_list;
      typedef std::list<Vector<Rational> > clist;	// constraint list
      typedef Array<Matrix<Rational> > matrix_list;
      typedef Array<Graph<Undirected> > graph_list;
    }

/* converts a list to a matrix
 *
 */
Matrix<Rational> list2matrix(const vertex_list& v) {
	int dim=(v.begin())->size();
	Matrix<Rational> A(v.size(),dim);

	int i=0;
	for (vertex_list::const_iterator it=v.begin(); it!=v.end(); ++it, ++i) {
		A.row(i)=*it;
	}
	return A;
}

/* converts a list to a matrix
 *
 */
Matrix<Rational> list2matrix(const clist& v) {
	int dim=(v.begin())->size();
	Matrix<Rational> A(v.size(),dim);

	int i=0;
	for (clist::const_iterator it=v.begin(); it!=v.end(); ++it, ++i) {
		A.row(i)=*it;
	}
	return A;
}

/* LP solver uses cdd
 *
 * returns optimal solution
 * called by Adj
 */
Vector<Rational> solve_lp(const Matrix<Rational>& constraints, const Vector<Rational>& objective) {
   typedef cdd_interface::solver<Rational> Solver;
	Solver solver;
	Matrix<Rational> E;
    Solver::lp_solution S=solver.solve_lp(constraints, E, objective, 1);
    return S.second;
}

/* computes the minkowski sum of the points listed in components (and rows of polytopes)
 *  v = v(1)+v(2)+..+(k) where v(j) = P_j(comp(j))
 *
 * returns the sumvector v
 * called by initial, local_search and addition
 */
Vector<Rational> components2vector(const Array<int>& components, const matrix_list& polytopes){
	Vector<Rational> result(polytopes[0].row(0).size());
	int i=0;
	for(Array<int>::const_iterator it=components.begin(); it!=components.end(); ++it, ++i)	{
		result+=polytopes[i].row(*it);
	}
	result[0]=1;
	return result;
}

/* computes a canonical vector in the relative iteriour of the normal cone N(v,P)
 * (can also return a vector [v,w], where w is a ...)
 *
 * TODO: comment
 */
Vector<Rational> canonical_vector(const int k, const Array<int>& components, const matrix_list& polytopes, const graph_list& graphs, bool setc_st=false) {
	clist c;	// constraints
	for (int j=0; j<k; ++j) {
		for (Entire<Graph<>::adjacent_node_list>::const_iterator it=entire(graphs[j].adjacent_nodes(components[j])); !it.at_end(); ++it) {
			c.push_back(polytopes[j].row(*it)-polytopes[j].row(components[j]));
		}
	}

	// constraints of the form e_j(v_j,i)^T \lambda + \lambda_0 \leq 0
	Matrix<Rational> c1=list2matrix(c);

	int dim=c1.cols();
	Vector<Rational> c3=ones_vector<Rational>(c.size());
	Matrix<Rational> c2 = (c1 | -c3);

	// bounds for \lambda_i: -1\leq\lambda_i\leq 1
	Vector<Rational> b1(ones_vector<Rational>(dim));
	Matrix<Rational> b2(unit_matrix<Rational>(dim));
	Matrix<Rational> b= ((b1 | b2) / (b1 | -b2));

	Matrix<Rational> constraints(c2/b);
	Vector<Rational> obj(unit_vector<Rational>(dim+1,dim));	// objective function
	Vector<Rational> r = solve_lp(constraints, obj);

	if(setc_st){ 					//Variation of c_st TODO: verbessern
		Rational d=r[r.size()-1];	// the "minimal distance" of r to the bounding hyperplanes
		Vector<Rational> x=r.slice(0,r.size()-1);
		Vector<Rational> eps(x.size());
		eps[1]= d/2;
		for (int i=2; i<eps.size(); ++i)
			eps[i]=Rational(-1,i)*eps[i-1];
		return r.slice(0,r.size()-1)|r.slice(0,r.size()-1)+eps;
	}
	return r.slice(0,r.size()-1);
}

/* checks whether two Vectors are parallel
 *
 * returns boolean
 * called by local_search_compare and Adj
 */
bool parallel_edges(const Vector<Rational>& e1, const Vector<Rational>& e2) {
	int dim=e1.size();
	int j=1;
	bool quotient_found=0;
	Rational q=0;		// quotient e2[j]/e1[j] -- should be constant
	while (!quotient_found) {
		if (e1[j]==0) {
			if (e2[j]!=0)
				return 0;
		}
		else {
			q=e2[j]/e1[j];
			quotient_found=1;
		}
		++j;
	}
	for (j=1; j<dim; ++j) {
		if ((e1[j]) * q != e2[j])
			return 0;
	}
	return 1;
}

/* computes the normal Vector for the by [c,c_st] first intersectet hyperplane (part of Normalcone N(v,P))
 * direction in which f(v) lies
 *
 * returns the normalvector
 * called by local_search and local_search_compare
 */
Vector<Rational> search_direction(const int k, const Vector<Rational>& c_st1, const Vector<Rational>& c_st2, const Array<int>& components, const matrix_list& polytopes, const graph_list& graphs) {
	// computes canonical vector for v
	Vector<Rational> c_st=c_st1;
	Vector<Rational> c = canonical_vector(k, components, polytopes, graphs);
	Rational best_theta=2;				// smallest theta value (so far)
	Vector<Rational> best_hyperplane;	// the hyperplane intersected first

	if(parallel_edges(c_st, c)){
		c_st=c_st2;
	}

	// goes through all (potential) neighbours of next
	for (int j=0; j<k; ++j) {
		for (Entire<Graph<>::adjacent_node_list>::const_iterator it=entire(graphs[j].adjacent_nodes(components[j])); !it.at_end(); ++it) {
			Vector<Rational> hyperplane = polytopes[j].row(*it)-polytopes[j].row(components[j]); //= e_j(next_j,i)
			// now compute theta such that c+theta(c_st-c) lies on the hyperplane
			Rational d = hyperplane* (c-c_st);
			if (d==0) {
				continue;	// the current hyperplane is parallel to c-c_st -> no intersection
			}
			Rational theta = (hyperplane*c) / d;
			if (0<=theta && 1>=theta && theta<best_theta) {
				best_theta=theta;
				best_hyperplane=hyperplane;
			}
		}
	}
	return best_hyperplane;
}

/*
 * TODO: comment
 */
Vector<Rational> local_search(const int k, const Vector<Rational>& c_st, const Vector<Rational>& c_st2, Array<int>& components, const matrix_list& polytopes, const graph_list& graphs) {

	//search normal Vector for the first intersectet hyperplane (part of Normalcone)
	Vector<Rational> hyperplane=search_direction(k, c_st, c_st2, components, polytopes, graphs);

	// now find the summands of f(v)
	for (int j=0; j<k; ++j) {
		for (Entire<Graph<>::adjacent_node_list>::const_iterator it=entire(graphs[j].adjacent_nodes(components[j])); !it.at_end(); ++it) {
			if (parallel_edges(hyperplane, polytopes[j].row(*it) - polytopes[j].row(components[j]))) {
				components[j] = *it;
				break;							// e_j(v_j,i) is in the right direction, so no other edge in P_j could be
			}
		}
	}
	return components2vector(components, polytopes);
}

/* checks whether f(next)==v, quicker variant of localsearch
 *
 * subroutins: parallel_edges and search_hyperplane
 * called by addition
 * TODO: kann verbessert werden in dem die richtung von next-v genutzt wird um ein erstes theta vorzugeben
 */
bool local_search_compare(const int k, const Vector<Rational>& v_st, const Vector<Rational>& c_st, const Vector<Rational>& c_st2, const Vector<Rational>& v, const Vector<Rational>& next, const Array<int>& next_components, const matrix_list& polytopes, const graph_list& graphs) {
	if(next==v_st){
		return 0;
	}
	//search normal Vector for the first intersectet hyperplane (part of Normalcone)
	Vector<Rational> hyperplane=search_direction(k, c_st, c_st2, next_components, polytopes, graphs);
	// now v == f(next) iff it lies in the right direction
	return parallel_edges(hyperplane,next-v);
}

/* searchs a adjacent edge of v in G(P) in direction e_s(v_s,r)
 *
 * returns 0 if there is no edge or the index (s,r) is not the smallest for this direction
 * subroutins: parallel_edges, list2matrix and solve_lp
 * called by addition
 */
bool Adj(const int k, Array<int>& next_components, const Array<int>& components, const int s, const int r, const graph_list& graphs, const matrix_list& polytopes){
	Vector<Rational> e_s=polytopes[s].row(r) - polytopes[s].row(components[s]); // =e_s(v_s,r)
	Vector<Rational> e_j;														// =e_j(v_j,i)
	clist ineq_list;															// list of ineq. (3.6) e_j(v_j,i)^T lambda >=0
	ineq_list.push_back(-e_s);													// and e_s(v_s,r)^T lambda < 0
	next_components=components;
	for(int j=0;j<k;++j){
		for (Entire<Graph<>::adjacent_node_list>::const_iterator it = entire(graphs[j].adjacent_nodes(components[j])); !it.at_end(); ++it) {
			e_j=polytopes[j].row(*it) - polytopes[j].row(components[j]);
			if(parallel_edges(e_s, e_j)){
				// returns 0 if there is an index j smaller than s in same direction
				if(j<s){
					return 0;}
				next_components[j]=*it;
			}else{
				ineq_list.push_back(e_j);
			}
		}
	}
	// add another variable mu in the first inequality
	// (the ineq will be strict iff mu is positive)
	Matrix<Rational> m = ( (list2matrix(ineq_list) ) | -unit_vector<Rational>(ineq_list.size(), 0) );
	int l=m.cols()-1;
	// one more ineq to bound lambda from above
	Vector<Rational> b = unit_vector<Rational>(m.cols(),0);
	b[l]=-1;
	Matrix<Rational> d= (m / b);
	// maximize lambda
	Vector<Rational> obj = unit_vector<Rational>(m.cols(),l);
	Vector<Rational> opt = solve_lp(d,obj);
	if (opt[l]==0)
		return 0;
	else {
		return 1;
	}
}

/* Reverse Search Algorithm according to Komei Fukuda
 * (see also MinkowskiAddition in 'From the Zonotope Construction to the Minkowski Addition of Convex Polytopes')
 *
 * subroutins: Adj, local_search, local_search_compare and components2vector
 * called by minkowski_sum_fukuda
 */
void addition(const int k, const Vector<Rational>& v_st, const Vector<Rational>& c_st, const Vector<Rational>& c_st2, vertex_list& vertices, Array<int>& components, const graph_list& graphs, matrix_list& polytopes){
	Vector<Rational> next;
	Array<int> next_components(k), u_components(k);
	Vector<Rational> v = v_st;
	int j=0;  																								// iterates through all input-polytopes
	Entire<Graph<>::adjacent_node_list>::const_iterator it=entire(graphs[j].adjacent_nodes(components[j]));	//iterates through all neighbors in P_j
	vertices.insert(v_st);

	while(j!=k or v!=v_st){
		while (j<k){
			while(!it.at_end()){
				if(Adj(k, next_components, components, j, *it, graphs, polytopes)){			 					// (r1)
					next=components2vector(next_components, polytopes);
					if(local_search_compare(k, v_st, c_st, c_st2, v, next, next_components, polytopes, graphs)){// (r2), f(next)==v
						//reverse traverse:
						components = next_components;
						v = next;
						j=0;
						it=entire(graphs[j].adjacent_nodes(components[j]));
						--it;
						vertices.insert(v);
					}
				}
				++it;
			}
			++j;
			if(j!=k){
				it=entire(graphs[j].adjacent_nodes(components[j]));
			}
		}
		if(v!=v_st){
			//forward traverse:
			u_components=components;										//(f1)
			v=local_search(k, c_st, c_st2, components, polytopes, graphs);	//v=f(v)
			//restore (j,it) s.t. Adj(v,(j,*it))=u:							//(f2)
			j=0;
			while(polytopes[j].row(u_components[j])==polytopes[j].row(components[j])){
				++j;
			}
			it=entire(graphs[j].adjacent_nodes(components[j]));
			while(*it!=u_components[j]){
				++it;
			}
			if(it.at_end()){
				++j;
				if(j!=k){
					it=entire(graphs[j].adjacent_nodes(components[j]));
				}
			}else{
				++it;
			}
		}
	}
}

/* computes the maximal face of polytope
 * (the code is essentially borrowed from apps/polytope/src/pseudo_simplex.cc)
 *
 * TODO: comment
 */
Set<int> find_max_face(const Matrix<Rational>& V, const Graph<Undirected>& G, const Vector<Rational>& objective) {

   NodeMap<Undirected, bool> visited(G,false);

   int current_vertex=0;

   Rational opt=objective * V[current_vertex];
   Set<int> optimal_face=scalar2set(current_vertex);
   visited[current_vertex]=true;

   bool better;

   do {
      better = false;
      // steepest ascent/descent
      for (Entire<Graph<>::out_edge_list>::const_iterator v=entire(G.out_edges(current_vertex));  !v.at_end();  ++v) {
         const int neighbor=v.to_node();
         if (visited[neighbor])  // this neighbor can't be better
            continue;

		visited[neighbor]=true;
		const Rational value = objective * V[neighbor];
		if (value > opt) { // this one is better
		   current_vertex = neighbor;
		   opt = value;
		   optimal_face = scalar2set(current_vertex);
		   better = true;
		} else if (value==opt) { // also belongs to optimal_face
		   optimal_face += neighbor;
		}
      }
   } while (better);

  std::list<int> optimal_vertices(optimal_face.begin(), optimal_face.end());

  while (!optimal_vertices.empty()) {
	 current_vertex = optimal_vertices.front();
	 optimal_vertices.pop_front();

	 for (Entire<Graph<>::out_edge_list>::const_iterator v=entire(G.out_edges(current_vertex));  !v.at_end();  ++v) {
		const int neighbor=v.to_node();
		if (!visited[neighbor]) {
		   visited[neighbor]=true;
		   if (!is_zero(V(neighbor,0)) && objective * V[neighbor] == opt) {
			  optimal_face += neighbor;
			  optimal_vertices.push_back(neighbor);
		   }
		}
	 }
  }

   return optimal_face;
}

/* decides whether row a is lager than row b, compared by lexicographical order
 *
 * returns a or b
 * called by initial
 */
int lex_max(const int a, const int b, const Matrix<Rational>& mat){
	Vector<Rational> v =mat.row(a)-mat.row(b);;
	for (Vector<Rational>::iterator it=v.begin(); it!=v.end(); ++it){
		if(*it>0)
			return a;
	}
	return b;
}

/* initial graphs, polytopes, components, v_st and c_st (c_st2)
 * using Array<perl::Object> summands (and k)
 *
 * and subroutins find_max_face, lex_max, components2vector and canonical_vector
 * called by minkowski_sum_fukuda
 */
void initial(const Array<perl::Object>& summands, const int k, graph_list& graphs, matrix_list& polytopes, Array<int>& components, Vector<Rational>& v_st, Vector<Rational>& c_st, Vector<Rational>& c_st2){
	//initial graphs and polytopes:
	int j=0;
	for (Array<perl::Object>::const_iterator it=summands.begin(); it!=summands.end(); ++it, ++j) {
	  	const Matrix<Rational> m=it->give("VERTICES");
		polytopes[j]=m;
		const Graph<Undirected> graph=it->give("GRAPH.ADJACENCY");
		graphs[j]=graph;
	}

	//initial components and v_st:
	Vector<Rational> obj(polytopes[0].row(0).size()); //LP Vector
	for (Vector<Rational>::iterator it=obj.begin(); it!=obj.end(); ++it)
		*it=1; 																//TODO: good direction
	for (int i=0; i<k; ++i) {
		Set<int> max_face = find_max_face(polytopes[i], graphs[i], obj);	//solves LP over P_j
		while(max_face.size()>1){											//(is the face more than a point (TODO: faster)
			Set<int>::iterator SetIterator=max_face.begin();
			++SetIterator;
			max_face.erase(lex_max(*(max_face.begin()),*SetIterator, polytopes[i])); //delete some, choosen by lex. order)
		}
		components[i]=*(max_face.begin());					 				//to find v_st
	}
	v_st=components2vector(components, polytopes);

	//initial c_st and c_st2 (canonical Vectors in N(P,v_st)):
	Vector<Rational> c = canonical_vector(k, components, polytopes, graphs, true);
	int size = c.size()/2;
	c_st = c.slice(0,size);
	c_st2= c.slice(size,size);
}

perl::Object minkowski_sum_fukuda(const Array<perl::Object>& summands){
	int k = summands.size();	// number of (input)polytopes
    Vector<Rational> v_st;  	// Root Verted v*
    Vector<Rational> c_st;  	// Canonical Vector c*
    Vector<Rational> c_st2; 	// alternativ for c* (if c is parallel to c_st)
	vertex_list vertices;		// Output = listed coordinates of the vertices of P, (the first vertex will be v*)
	Array<int> components(k);	// j-th entry will contain the number of a vertex in P_j  s.t. v = v(1)+v(2)+...+v(k) where v(j)= P_j(comp(j))
	graph_list graphs(k);		// stores all Graphs from the input Polytopes P_j
	matrix_list polytopes(k);	// stores matrices s.t. the i-th entry is a discribtion of P_j by vertices

	initial(summands, k, graphs, polytopes, components, v_st, c_st, c_st2);
	addition(k, v_st, c_st, c_st2, vertices, components, graphs, polytopes);

	Matrix<Rational> ver = list2matrix(vertices);   // converts vertex_list to Matrix
	perl::Object ret("Polytope<Rational>");
	ret.take("VERTICES")<<ver;
	return ret;
}

UserFunctionTemplate4perl("# @category Producing a new polyhedron from others", "minkowski_sum_fukuda(Polytope<Rational> +)");

}}
