/* Copyright (c) 1997-2010
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Darmstadt, Germany)
   http://www.polymake.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
--------------------------------------------------------------------------------
   $Project: polymake $$Id: polarize.cc 10414 2011-11-17 14:00:56Z paffenho $
*/

#include "polymake/client.h"
#include "polymake/Matrix.h"
#include "polymake/Array.h"
#include "polymake/IncidenceMatrix.h"
#include "polymake/linalg.h"

namespace polymake { namespace polytope {

template <typename Scalar>
perl::Object polarize(perl::Object p_in, perl::OptionSet options)
{
   bool some_section_read = false;
   const bool no_coordinates=options["noc"];
   const bool isCone = !p_in.isa("Polytope");

   perl::ObjectType t=p_in.type();
   perl::Object p_out(t);
   if ( isCone ) 
      p_out.set_description() << "Cone dualized from " << p_in.name() << endl;
   else
      p_out.set_description() << "Polytope polarized from " << p_in.name() << endl;

   if (no_coordinates || p_in.exists("RAYS_IN_FACETS")) {
      const IncidenceMatrix<> VIF=p_in.give("RAYS_IN_FACETS");
      p_out.take("VERTICES_IN_FACETS") << T(VIF);
      some_section_read = true;
   }

   if (!no_coordinates) {
      const int  ambient_dim=p_in.give("CONE_AMBIENT_DIM");
      
      if ( !isCone ) {
         Vector<Scalar> u = unit_vector<Scalar>(ambient_dim,0);
         const bool is_centered=p_in.CallPolymakeMethod("contains",u);
         if (!is_centered)
            throw std::runtime_error("polarize: polytope must contain the origin");
      }

      Matrix<Scalar> ineq, eq, pts, lin;
      std::string got_property;

      // at least one of the following checks must be true, 
      // otherwise the input does not define a valid geometric Cone or Polytope
      if (p_in.lookup_with_property_name("RAYS | INPUT_RAYS", got_property) >> ineq) {
         p_out.take( got_property=="RAYS" || got_property=="VERTICES" ? "FACETS" : "INEQUALITIES") << ineq;
         some_section_read = true;
      }

      if (p_in.lookup_with_property_name("FACETS | INEQUALITIES", got_property) >> pts) {
         p_out.take( got_property=="FACETS" ? "RAYS" : "INPUT_RAYS") << pts;
         some_section_read = true;
      }

      // the following properties may exist, and if so, are essential to describe the polyhedron
      // but they are not valid without one of the provious properties
      if (p_in.lookup_with_property_name("LINEALITY_SPACE | INPUT_LINEALITY", got_property) >> eq)
         p_out.take( got_property=="LINEALITY_SPACE" ? "LINEAR_SPAN" : "EQUATIONS") << eq;
      if (p_in.lookup_with_property_name("LINEAR_SPAN | EQUATIONS", got_property) >> lin)
         p_out.take( got_property=="LINEAR_SPAN" || got_property=="AFFINE_HULL" ? "LINEALITY_SPACE" : "INPUT_LINEALITY") << lin;

      // let's see if we produced something at all
      if (!some_section_read)
         throw std::runtime_error("polarize: no useful coordinate section found in input");

      // if p was obtained by some transformation,
      // then the reverse transformation also makes sense for the polar
      Matrix<Scalar> tau;
      if (p_in.get_attachment("REVERSE_TRANSFORMATION") >> tau)
         p_out.attach("REVERSE_TRANSFORMATION") << T(inv(tau));
        
      p_out.take("CONE_AMBIENT_DIM") << ambient_dim;
      int ldim, cdim;
      if (p_in.lookup("LINEALITY_DIM") >> ldim ) 
         p_out.take("CONE_DIM") << ambient_dim-ldim;      
      if (p_in.lookup("CONE_DIM") >> cdim ) 
         p_out.take("LINEALITY_DIM") << ambient_dim-cdim;      
   }

   Array<std::string> labels;
   if (p_in.lookup("RAY_LABELS") >> labels)
      p_out.take("FACET_LABELS") << labels;
   if (p_in.lookup("FACET_LABELS") >> labels)
      p_out.take("RAY_LABELS") << labels;

   return p_out;
}

UserFunctionTemplate4perl("# @category Transforming a polyhedron"
                          "# Given a bounded, centered, and full-dimensional polytope //P//, produce its polar,"
                          "# that is, polar with respect to the standard Euclidean scalar product."
                          "# Note that the definition of the polar has changed after version 2.10: "
                          "# the polar is reflected in the origin to conform with cone dualization"
                          "# @param Cone C"
                          "# @option Bool noc only combinatorial information is handled"
                          "# @return Cone",
                          "polarize<Scalar> (Cone<Scalar>, { noc => 0 })");
} }

// Local Variables:
// mode:C++
// c-basic-offset:3
// indent-tabs-mode:nil
// End:
