#include "polymake/client.h"
#include "polymake/Vector.h"
#include "polymake/Matrix.h"
#include "polymake/Rational.h"
#include "polymake/linalg.h"
#include "polymake/polytope/lrs_interface.h"

namespace polymake { namespace polytope {

void is_vertex_sub(const Vector<Rational>& q, const Matrix<Rational>& points, bool& answer, Vector<Rational>& sep_hyp)
{
   /*construction of LP according to cdd redundancy check for points, see
     http://www.ifor.math.ethz.ch/~fukuda/polyfaq/node22.html#polytope:Vredundancy
    */
   Matrix<Rational> extension(zero_vector<Rational>(points.rows())/ones_vector<Rational>(points.rows()));
   Matrix<Rational> ineqs(T(extension)|-points.minor(All,range(1,points.cols()-1)));//z^tp_i-z_0<=0; CAUTION: p_i is affine part of i-th point! 
   Vector<Rational> prev_unbound(ones_vector<Rational>(2)|-q.slice(1)); //z^tq-z_0<=1, prevents unboundedness
   ineqs/=prev_unbound;
   Matrix<Rational> aff_hull=null_space(points/q);
   Matrix<Rational> extension2(aff_hull.rows(),2);
   Matrix<Rational> eqs(extension2|-aff_hull.minor(All, range(1,aff_hull.cols()-1)));
   Vector<Rational> obj(zero_vector<Rational>(1)|-ones_vector<Rational>(1)|q.slice(1)); //z^tq-z_0

   lrs_interface::solver S;
   /// @retval first: objective value, second: solution
   typedef std::pair<Rational, Vector<Rational> > lp_solution;
   lp_solution sol=S.solve_lp(ineqs, eqs, obj, true);
   Rational obj_val=sol.first;
   Vector<Rational> res(sol.second);
   //H: z^tx=z_0, i.e., z_0-z^tx=0
   sep_hyp=-res.slice(1); //sep_hyp*point should be >= 0
   sep_hyp[0]=res[1]; 
   if(sol.first > 0){ //q non-red. <=> obj_val > 0
      answer=true;
   } else {
      answer=false;
   }
}


bool is_vertex(const Vector<Rational>& q, const Matrix<Rational>& points)
{
   bool answer;
   Vector<Rational> sep_hyp(points.cols());
   is_vertex_sub(q, points, answer, sep_hyp);
   return answer;
}


perl::ListReturn separating_hyperplane(const Vector<Rational>& q, const Matrix<Rational>& points)
{
   bool answer;
   Vector<Rational> sep_hyp(points.cols());
   is_vertex_sub(q, points, answer, sep_hyp);

   perl::ListReturn result;
   result << answer
          << sep_hyp;
   return result;
}


UserFunction4perl("# @category Other"
                  "# Checks whether a given point //q// is a vertex of the polytope P generated by"
                  "# //q// and a set of other points //points// via solving a suitable LP"
                  "# (compare cdd redundancy check). Works without knowing the facets of P!"
                  "# @param Vector q the vertex (candidate) which is to be separated from //points//"
                  "# @param Matrix points the points from which //q// is to be separated"
                  "# @return Bool 'true' if //q// is a vertex",
                  &is_vertex,"is_vertex(Vector, Matrix)");

UserFunction4perl("# @category Other"
                  "# Computes (the normal vector of) a hyperplane which separates a given point //q//"
                  "# from //points// via solving a suitable LP. The scalar product of the normal vector"
                  "# of the separating hyperplane and a point in //points// is greater or equal than 0"
                  "# (same behavior as for facets!)."
                  "# If //q// is not a vertex of P=conv(//points//,//q//),"
                  "# the function returns a zero vector and sets //answer// to 'false'."
                  "# Works without knowing the facets of P!" 
                  "# @param Vector q the vertex (candidate) which is to be separated from //points//"
                  "# @param Matrix points the points from which //q// is to be separated"
                  "# @return ListReturn (Bool answer, Vector sep_hyp)",
                  &separating_hyperplane,"separating_hyperplane(Vector, Matrix)");

} }




// Local Variables:
// mode:C++
// c-basic-offset:3
// indent-tabs-mode:nil
// End:
